import unittest
from copy import copy
from typing import Iterable

from hgijson import JsonPropertyMapping, MappingJSONEncoderClassBuilder, MappingJSONDecoderClassBuilder


class _NoPropertyAccess:
    def __init__(self, configurations: Iterable):
        self._configurations = configurations

    def get_configurations(self) -> Iterable:
        return copy(self._configurations)


_no_property_access_mappings = [
    JsonPropertyMapping(json_property_name="configurations", object_constructor_parameter_name="configurations",
                        object_property_getter=lambda obj: obj.get_configurations())
]
_NoPropertyAccessJSONEncoder = MappingJSONEncoderClassBuilder(_NoPropertyAccess, _no_property_access_mappings).build()
_NoPropertyAccessJSONDecoder = MappingJSONDecoderClassBuilder(_NoPropertyAccess, _no_property_access_mappings).build()


class TestNoPropertyAccess(unittest.TestCase):
    """
    Test no object property access support.
    """
    def test_no_property_access(self):
        configurations = [1, 2, 3]
        obj = _NoPropertyAccess(configurations)
        json_as_dict = _NoPropertyAccessJSONEncoder().default(obj)
        self.assertEqual({"configurations": configurations}, json_as_dict)


if __name__ == "__main__":
    unittest.main()



import json
import unittest

from hgijson.json_converters._converters import json_encoder_to_serializer, json_decoder_to_deserializer
from hgijson.serialization import Deserializer, Serializer
from hgijson.tests.json_converters._helpers import create_simple_model_with_json_representation
from hgijson.tests.json_converters._serializers import BasicSimpleModelJSONEncoder, BasicSimpleModelJSONDecoder


class TestConverters(unittest.TestCase):
    """
    Tests for `json_encoder_to_serializer` and `json_decoder_to_deserializer`.
    """
    def setUp(self):
        self.simple_model = create_simple_model_with_json_representation()[0]
        self.simple_model_as_json = create_simple_model_with_json_representation()[1]

    def test_json_encoder_to_serializer(self):
        serializer_cls = json_encoder_to_serializer(BasicSimpleModelJSONEncoder)
        serializer = serializer_cls()   # type: Serializer
        self.assertIsInstance(serializer, Serializer)

        serialized = serializer.serialize(self.simple_model)
        encoded = BasicSimpleModelJSONEncoder().default(self.simple_model)
        self.assertDictEqual(serialized, encoded)

    def test_json_decoder_to_deserializer(self):
        deserializer_cls = json_decoder_to_deserializer(BasicSimpleModelJSONDecoder)
        deserializer = deserializer_cls()   # type: Deserializer
        self.assertIsInstance(deserializer, Deserializer)

        deserialized = deserializer.deserialize(self.simple_model_as_json)
        decoded = BasicSimpleModelJSONDecoder().decode(json.dumps(self.simple_model_as_json))
        self.assertEqual(deserialized, decoded)


if __name__ == "__main__":
    unittest.main()



from json import JSONEncoder

from hgijson.json_converters.automatic import _RegisteredTypeJSONEncoder
from hgijson.tests._models import BaseModel


class StubModel(BaseModel):
    """
    Stub `Model`.
    """


class StubRegisteredTypeJSONEncoder(_RegisteredTypeJSONEncoder):
    """
    Stub `_RegisteredTypeJSONEncoder`.
    """
    def _get_json_encoders_for_type(self) -> JSONEncoder:
        """ Unused - to satisfy the interface only. """



from abc import abstractmethod, ABCMeta
from json import JSONDecoder

from hgijson.custom_types import SerializableType, PrimitiveJsonType


class ParsedJSONDecoder(JSONDecoder, metaclass=ABCMeta):
    """
    Decoder of JSON parsed from a string into primitive Python objects.
    """
    @abstractmethod
    def decode_parsed(self, parsed_json: PrimitiveJsonType) -> SerializableType:
        """
        Decodes the given JSON, represented as primitive Python objects.
        :param parsed_json: the JSON
        :return: the decoded object
        """



from typing import Any, Iterable, Sequence, List, Type

from hgijson.json_converters.models import JsonPropertyMapping
from hgijson.serialization import Serializer, Deserializer, PropertyMapping
from hgijson.serializers import PrimitiveSerializer, PrimitiveDeserializer
from hgijson.tests._models import SimpleModel, ComplexModel


def get_simple_model_property_mappings() -> Sequence[PropertyMapping]:
    return [
        JsonPropertyMapping("serialized_a", "a"),
        JsonPropertyMapping("serialized_b", "b", "constructor_b")
    ]


def get_complex_model_property_mappings() -> Sequence[PropertyMapping]:
    return [
        JsonPropertyMapping("serialized_b", "b", "constructor_b"),
        JsonPropertyMapping("serialized_c", "c"),
        PropertyMapping(serialized_property_getter=lambda obj_as_dict: obj_as_dict["serialized_d"],
                        serialized_property_setter=lambda obj_as_dict, value: obj_as_dict.__setitem__(
                            "serialized_d", value),
                        object_property_getter=lambda obj: obj.d,
                        object_property_setter=lambda obj, value: setattr(obj, "d", value),
                        serializer_cls=SimpleModelSerializer,
                        deserializer_cls=SimpleModelDeserializer),
        JsonPropertyMapping("serialized_e", "e"),
        JsonPropertyMapping("serialized_f", "f"),
        JsonPropertyMapping("serialized_g", "g"),
        JsonPropertyMapping("serialized_h", "h"),
        JsonPropertyMapping("serialized_i", "i", collection_factory=set, collection_iter=iter),
    ]


class SimpleModelSerializer(Serializer):
    """
    Serializer for `SimpleModel`.
    """
    def __init__(self, custom_mappings: Iterable[PropertyMapping]=None):
        if custom_mappings is None:
            custom_mappings = get_simple_model_property_mappings()
        super().__init__(custom_mappings)

    def _create_serialized_container(self) -> Any:
        return {}

    def _create_serializer_of_type(self, serializer_type: Type) -> Serializer:
        return PrimitiveSerializer()


class ComplexModelSerializer(Serializer):
    """
    Serializer for `ComplexModel`.
    """
    def __init__(self, custom_mappings: Iterable[PropertyMapping]=None):
        if custom_mappings is None:
            custom_mappings = list(get_simple_model_property_mappings()) + list(get_complex_model_property_mappings())
        super().__init__(custom_mappings)

    def _create_serialized_container(self) -> Any:
        return {}

    def _create_serializer_of_type(self, serializer_type: Type) -> Serializer:
        if serializer_type == SimpleModelSerializer:
            return SimpleModelSerializer()
        else:
            return PrimitiveSerializer()


class SimpleModelDeserializer(Deserializer):
    """
    Deserializer for `SimpleModel`.
    """
    def __init__(self, custom_mappings: Iterable[PropertyMapping]=None):
        if custom_mappings is None:
            custom_mappings = get_simple_model_property_mappings()
        super().__init__(custom_mappings, SimpleModel)

    def _create_deserializer_of_type(self, deserializer_type: Type):
        return PrimitiveDeserializer()


class ComplexModelDeserializer(Deserializer):
    """
    Deserializer for `ComplexModel`.
    """
    def __init__(self, custom_mappings: Iterable[PropertyMapping]=None):
        if custom_mappings is None:
            custom_mappings = list(get_simple_model_property_mappings()) + list(get_complex_model_property_mappings())
        super().__init__(custom_mappings, ComplexModel)

    def _create_deserializer_of_type(self, deserializer_type: Type):
        if deserializer_type == SimpleModelDeserializer:
            return SimpleModelDeserializer()
        else:
            return PrimitiveDeserializer()



import json
from abc import ABCMeta, abstractmethod
from json import JSONDecoder, JSONEncoder
from typing import Optional, Union, Callable, Type

from hgijson.json_converters.interfaces import ParsedJSONDecoder
from hgijson.serialization import Deserializer, Serializer
from hgijson.custom_types import PrimitiveJsonType, PrimitiveUnionType, SerializableType


class _JSONEncoderAsSerializer(Serializer, metaclass=ABCMeta):
    """
    JSON encoder wrapped to work as a `Serializer`.
    """
    @property
    @abstractmethod
    def encoder_type(self) -> Type[JSONEncoder]:
        """
        Gets the `JSONEncoder` type that is used by this serialiser.
        :return: the encoder type (must be a subclass of `JSONEncoder`)
        """

    def __init__(self, *args, **kwargs):
        super().__init__([])
        self._encoder = self.encoder_type(*args, **kwargs)

    def serialize(self, serializable: Optional[SerializableType]) -> PrimitiveUnionType:
        if type(self._encoder) == JSONEncoder:
            # Have to load from string to more rich representation - not possible to stop `encode` going to string :/
            return json.loads(self._encoder.encode(serializable))
        else:
            return self._encoder.default(serializable)

    def _create_serializer_of_type(self, serializer_type: Type[Serializer]) -> None:
        """
        Unused - implemented to satisfy the interface only.
        """

    def _create_serialized_container(self) -> None:
        """
        Unused - implemented to satisfy the interface only.
        """


class _JSONDecoderAsDeserializer(Deserializer, metaclass=ABCMeta):
    """
    JSON decoder wrapped to work as a `Deserializer`.
    """
    @property
    @abstractmethod
    def decoder_type(self) -> Type[JSONDecoder]:
        """
        Gets the `JSONDecoder` type that is used by this deserialiser.
        :return: the decoder type (must be a subclass of `JSONDecoder`)
        """

    def __init__(self, *args, **kwargs):
        super().__init__([], self.decoder_type)
        self._decoder = self.decoder_type(*args, **kwargs)

    def deserialize(self, deserializable: PrimitiveJsonType) -> Optional[SerializableType]:
        if not isinstance(self._decoder, ParsedJSONDecoder):
            # Decode must take a string (even though we have a richer representation) :/
            json_as_string = json.dumps(deserializable)
            return self._decoder.decode(json_as_string)
        else:
            # Optimisation - no need to convert our relatively rich representation into a string (just to turn it back
            # again!)
            return self._decoder.decode_parsed(deserializable)

    def _create_deserializer_of_type(self, deserializer_type: Type[JSONDecoder]) -> None:
        """
        Unused - implemented to satisfy the interface only.
        """


def json_encoder_to_serializer(encoder_cls: Union[Type[JSONEncoder], Callable[[], Type[JSONEncoder]]]) \
        -> Type[Serializer]:
    """
    Converts a `JSONEncoder` class into an equivalent `Serializer` class.
    :param encoder_cls: the encoder class type or a function that returns the type
    :return: the equivalent `Serializer` class
    """
    name = encoder_cls.__name__ if isinstance(encoder_cls, type) else "%sLambdaTypeReturn" % id(encoder_cls)
    return type(
        "%sAsSerializer" % name,
        (_JSONEncoderAsSerializer,),
        {
            "encoder_type": property(lambda self: encoder_cls if isinstance(encoder_cls, type) else encoder_cls())
        }
    )


def json_decoder_to_deserializer(decoder_cls: Union[Type[JSONDecoder], Callable[[], Type[JSONDecoder]]]) \
        -> Type[Deserializer]:
    """
    Converts a `JSONDecoder` class into an equivalent `Deserializer` class.
    :param decoder_cls: the decoder class type or a function that returns the type
    :return: the equivalent `Deserializer` class
    """
    name = decoder_cls.__name__ if isinstance(decoder_cls, type) else "%sLambdaTypeReturn" % id(decoder_cls)
    return type(
        "%sAsDeserializer" % name,
        (_JSONDecoderAsDeserializer,),
        {
            "decoder_type": property(lambda self: decoder_cls if isinstance(decoder_cls, type) else decoder_cls())
        }
    )



import json
from datetime import datetime, timezone
from json import JSONDecoder, JSONEncoder
from typing import Any, TypeVar

from dateutil.parser import parser

from hgijson.json_converters.interfaces import ParsedJSONDecoder

ItemType = TypeVar("ItemType")


class StrJSONEncoder(JSONEncoder):
    """
    JSON encoder from any type that implements `__str__` to its string representation.
    """
    def default(self, to_encode: Any) -> str:
        return str(to_encode)


class StrJSONDecoder(JSONDecoder):
    """
    JSON decoder for strings.
    """
    def decode(self, to_decode: str, **kwargs) -> str:
        return str(to_decode)


class IntJSONEncoder(JSONEncoder):
    """
    JSON encoder from any type that implements `__int__`  to an integer.
    """
    def default(self, to_encode: Any) -> str:
        return int(to_encode)


class IntJSONDecoder(JSONDecoder):
    """
    JSON decoder for integers.
    """
    def decode(self, to_decode: str, **kwargs) -> int:
        return int(to_decode)


class FloatJSONEncoder(JSONEncoder):
    """
    JSON encoder from any type that implements `__float__`  to a float.
    """
    def default(self, to_encode: Any) -> str:
        return float(to_encode)


class FloatJSONDecoder(JSONDecoder):
    """
    JSON decoder for floats.
    """
    def decode(self, to_decode: str, **kwargs) -> str:
        return float(to_decode)


class DatetimeISOFormatJSONEncoder(JSONEncoder):
    """
    JSON encoder for datetime to ISO 8601 format.
    """
    def default(self, to_encode: datetime) -> str:
        return to_encode.isoformat()


class DatetimeISOFormatJSONDecoder(ParsedJSONDecoder):
    """
    JSON decoder for datetime as ISO 8601 formatted string.
    """
    _DATE_PARSER = parser()

    def decode(self, to_decode: str, **kwargs) -> str:
        return self.decode_parsed(json.loads(to_decode))

    def decode_parsed(self, parsed_json: str) -> str:
        return DatetimeISOFormatJSONDecoder._DATE_PARSER.parse(parsed_json)


class DatetimeEpochJSONEncoder(JSONEncoder):
    """
    JSON encoder for datetime to seconds since the epoch (1970-01-01). If the datetime has microsecond precision, it
    will be rounded to the nearest corresponding second since the epoch.
    """
    def default(self, to_encode: datetime) -> int:
        return int(to_encode.timestamp())


class DatetimeEpochJSONDecoder(JSONDecoder):
    """
    JSON decoder for datetime as seconds since the epoch (1970-01-01).
    """
    def decode(self, to_decode: str, **kwargs) -> datetime:
        return datetime.fromtimestamp(int(to_decode), timezone.utc)



from typing import Dict

from hgijson.serialization import Serializer, Deserializer


class JsonObjectSerializer(Serializer):
    """
    JSON serializer for models represented by {}.
    """
    _JSON_ENCODER_ARGS = []
    _JSON_ENCODER_KWARGS = {}

    def _create_serializer_of_type(self, serializer_type: type) -> Serializer:
        return serializer_type(*self._JSON_ENCODER_ARGS, **self._JSON_ENCODER_KWARGS)

    def _create_serialized_container(self) -> Dict:
        return {}


class JsonObjectDeserializer(Deserializer):
    """
    JSON deserializer for models represented by {}.
    """
    _JSON_ENCODER_ARGS = []
    _JSON_ENCODER_KWARGS = {}

    def _create_deserializer_of_type(self, deserializer_type: type) -> Deserializer:
        return deserializer_type(*self._JSON_ENCODER_ARGS, **self._JSON_ENCODER_KWARGS)



from hgijson.json_converters.primitive import IntJSONEncoder, StrJSONEncoder, IntJSONDecoder, FloatJSONEncoder, \
    FloatJSONDecoder, DatetimeISOFormatJSONEncoder, DatetimeISOFormatJSONDecoder, DatetimeEpochJSONEncoder, \
    DatetimeEpochJSONDecoder, ItemType, StrJSONDecoder

from hgijson.json_converters.builders import MappingJSONDecoderClassBuilder, MappingJSONEncoderClassBuilder

from hgijson.json_converters.models import JsonPropertyMapping



import json
import unittest
from json import JSONEncoder

from hgijson.json_converters.automatic import AutomaticJSONEncoderClassBuilder
from hgijson.tests._stubs import StubModel, StubRegisteredTypeJSONEncoder


class TestRegisteredTypeJSONEncoder(unittest.TestCase):
    """
    Tests for `_RegisteredTypeJSONEncoder`.
    """
    def test_default_when_unknown(self):
        self.assertRaises(TypeError, json.dumps, StubModel(), cls=StubRegisteredTypeJSONEncoder)

    def test_default_when_standard_type(self):
        dict = {1: 2, 3: []}
        self.assertEqual(json.dumps(dict, cls=StubRegisteredTypeJSONEncoder), json.dumps(dict))


class TestAutomaticJSONEncoderClassBuilder(unittest.TestCase):
    """
    Tests for `AutomaticJSONEncoderClassBuilder`.
    """
    def setUp(self):
        self.encoder_builder = AutomaticJSONEncoderClassBuilder()

    def test_get_json_encoders_for_type_if_not_known(self):
        self.assertIsNone(self.encoder_builder.get_json_encoders_for_type(JSONEncoder))

    def test_get_json_encoders_for_type_has_encoders_for_standard_types(self):
        repeat_to_test_reset = True
        types_to_check = [dict, list, tuple, str, int, float, bool, type(None)]
        while repeat_to_test_reset:
            for type_to_check in types_to_check:
                self.assertTrue(issubclass(self.encoder_builder.get_json_encoders_for_type(type_to_check), JSONEncoder))
            self.encoder_builder.reset_registered_json_encoders()
            repeat_to_test_reset = False

    def test_get_json_encoders_for_type_if_registered(self):
        self.encoder_builder.register_json_encoder(JSONEncoder, JSONEncoder)
        self.assertEqual(self.encoder_builder.get_json_encoders_for_type(JSONEncoder), JSONEncoder)

    def test_reset_registered_json_encoders(self):
        self.encoder_builder.register_json_encoder(JSONEncoder, JSONEncoder)
        self.encoder_builder.reset_registered_json_encoders()
        self.assertIsNone(self.encoder_builder.get_json_encoders_for_type(JSONEncoder))

    def test_build_with_no_registered_encoders(self):
        Encoder = self.encoder_builder.build()
        self.assertTrue(issubclass(Encoder, JSONEncoder))
        # Should work with types supported by default
        test_object = {1: 2}
        self.assertEqual(json.dumps(test_object, cls=Encoder), json.dumps(test_object))
        self.assertRaises(TypeError, json.dumps, StubModel(), cls=Encoder)

    def test_build_with_registered_encoders(self):
        # Sanity check
        self.assertRaises(TypeError, json.dumps, StubModel(), cls=self.encoder_builder.build())

        expect_encode = "expected encoding"

        class StubModelJSONEncoder(JSONEncoder):
            def default(self, o):
                assert isinstance(o, StubModel)
                return expect_encode
        self.encoder_builder.register_json_encoder(StubModel, StubModelJSONEncoder)

        self.assertEqual(json.dumps(StubModel(), cls=self.encoder_builder.build()), json.dumps(expect_encode))

    def test_build_not_influenced_by_future_registrations(self):
        Encoder = self.encoder_builder.build()

        expect_encode = "expected encoding"

        class StubModelJSONEncoder(JSONEncoder):
            def default(self, o):
                assert isinstance(o, StubModel)
                return expect_encode
        self.encoder_builder.register_json_encoder(StubModel, StubModelJSONEncoder)

        self.assertRaises(TypeError, json.dumps, StubModel(), cls=Encoder)
        self.assertEqual(json.dumps(StubModel(), cls=self.encoder_builder.build()), json.dumps(expect_encode))


if __name__ == "__main__":
    unittest.main()



from abc import ABCMeta
from typing import Iterable, Tuple, List

from hgijson.json_converters._serialization import MappingJSONEncoder, MappingJSONDecoder, PropertyMapper
from hgijson.json_converters.models import JsonPropertyMapping


class _JSONSerializationClassBuilder(metaclass=ABCMeta):
    """
    Subclass of serialization class builders.
    """
    def __init__(self, target_cls: type=type(None), mappings: Iterable[JsonPropertyMapping]=(),
                 superclasses: Tuple=None):
        """
        Constructor.
        :param superclasses: the superclasses to which the serialization class should extend
        :param target_cls: the class that the builder targets
        :param mappings: mappings from JSON properties to object properties
        """
        self.superclasses = superclasses
        self.target_cls = target_cls
        self.mappings = mappings


def _get_all_property_mappings(encoder: MappingJSONEncoder, property_mappings: Iterable[JsonPropertyMapping],
                               superclasses: Tuple[PropertyMapper]) -> List[JsonPropertyMapping]:
    """
    Gets all of the property mappings from the given property mapper, considering the property mappings for self and the
    property mappings defined by the superclass.
    :param encoder: `self` when binded as class method
    :param property_mappings: mappings defined for the given encoder, excluding mappings defined by superclasses
    :param superclasses: superclasses of the given encoder. Property mappers in later superclasses may override the
    effects of property mappers defined by superclasses closer to the start of the list
    :return: all of the property mappings for the given encoder
    """
    mappings = []
    for superclass in superclasses:
        super_mappings = superclass._get_property_mappings(superclass)
        mappings.extend(super_mappings)

    # Add property mappings of own to end of the mappings list
    mappings.extend(property_mappings)

    # Note: It is very difficult to cull all property mappers that target the same properties, leaving only the ones
    # from the lowest class in the hierarchy. This is because such mappers may be encoded as functions. Given that such
    # overloading is unlikely to be used much and the cost of doing a mapping and then mapping again over the top of it
    # will likely be small, there will be no attempt of such a cull.
    return mappings


class MappingJSONEncoderClassBuilder(_JSONSerializationClassBuilder):
    """
    Builder for `MappingJSONEncoder` concrete subclasses.
    """
    def __init__(self, target_cls: type=type(None), mappings: Iterable[JsonPropertyMapping]=(),
                 superclasses: Tuple=(MappingJSONEncoder, )):
        super().__init__(target_cls, mappings, superclasses)

    def build(self) -> type:
        """
        Build a subclass of `MappingJSONEncoder`.
        :return: the built subclass
        """
        def _get_property_mappings(encoder: MappingJSONEncoder) -> List[JsonPropertyMapping]:
            return _get_all_property_mappings(encoder, self.mappings, self.superclasses)

        def get_serializable_cls(encoder: MappingJSONEncoder) -> type:
            return self.target_cls

        def default(encoder: MappingJSONEncoder, serializable):
            if serializable is None:
                # Fix for #18
                return None
            elif isinstance(serializable, List):
                # Fix for #8
                return [encoder.default(item) for item in serializable]
            else:
                # Sort subclasses so subclass' default method is called last
                superclasses_as_list = list(self.superclasses)
                superclasses_as_list.sort(key=lambda superclass: 1 if superclass == MappingJSONEncoder else -1)

                encoded_combined = {}
                for superclass in superclasses_as_list:
                    encoded = superclass.default(encoder, serializable)
                    encoded_combined.update(encoded)

                return encoded_combined

        return type(
            "%sDynamicMappingJSONEncoder" % self.target_cls.__name__,
            self.superclasses,
            {
                "_get_property_mappings": _get_property_mappings,
                "_get_serializable_cls": get_serializable_cls,
                "default": default
            }
        )


class MappingJSONDecoderClassBuilder(_JSONSerializationClassBuilder):
    """
    Builder for `MappingJSONDecoder` concrete subclasses.
    """
    def __init__(self, target_cls: type=type(None), mappings: Iterable[JsonPropertyMapping]=(),
                 superclasses: Tuple=(MappingJSONDecoder, )):
        super().__init__(target_cls, mappings, superclasses)

    def build(self) -> type:
        """
        Build a subclass of `MappingJSONDecoder`.
        :return: the built subclass
        """
        def _get_property_mappings(encoder: MappingJSONEncoder) -> List[JsonPropertyMapping]:
            return _get_all_property_mappings(encoder, self.mappings, self.superclasses)

        def get_deserializable_cls(decoder: MappingJSONDecoder) -> type:
            return self.target_cls

        return type(
            "%sDynamicMappingJSONDecoder" % self.target_cls.__name__,
            self.superclasses,
            {
                "_get_property_mappings": _get_property_mappings,
                "_get_deserializable_cls": get_deserializable_cls
            }
        )



import unittest

from hgijson.json_converters.models import JsonPropertyMapping


class TestPropertyMapping(unittest.TestCase):
    """
    Tests for `JsonPropertyMapping`.
    """
    def test_init_with_serialized_property_name_and_getter_and_setter(self):
        self.assertRaises(ValueError, JsonPropertyMapping, json_property_name="a",
                          json_property_getter=lambda: None, json_property_setter=lambda: None)

    def test_init_with_object_and_serialized_names(self):
        property_mapping = JsonPropertyMapping("serialized_a", "a", "constructor_a")
        self.assertIsNotNone(property_mapping.object_property_getter)
        self.assertIsNone(property_mapping.object_property_setter)
        self.assertIsNotNone(property_mapping.serialized_property_getter)
        self.assertIsNotNone(property_mapping.serialized_property_setter)
        self.assertIsNotNone(property_mapping.object_constructor_parameter_name)

    def test_json_property_getter_property(self):
        property_mapping = JsonPropertyMapping("a")
        getter = lambda: None
        property_mapping.json_property_getter = getter
        self.assertEqual(property_mapping.json_property_getter, getter)

    def test_json_property_setter_property(self):
        property_mapping = JsonPropertyMapping("a")
        setter = lambda: None
        property_mapping.json_property_setter = setter
        self.assertEqual(property_mapping.json_property_setter, setter)

    def test_json_property_setter_property_when_non_optional_property_is_missing(self):
        property_mapping = JsonPropertyMapping("a")
        self.assertRaises(KeyError, property_mapping.json_property_getter, {})


if __name__ == "__main__":
    unittest.main()



import unittest
from copy import deepcopy

from hgijson import JsonPropertyMapping, MappingJSONEncoderClassBuilder, MappingJSONDecoderClassBuilder
from hgijson.tests._models import BaseModel
from hgijson.tests.json_converters._helpers import EXAMPLE_VALUE_1, EXAMPLE_PROPERTY_1, EXAMPLE_PROPERTY_2, \
    EXAMPLE_VALUE_2, EXAMPLE_PROPERTY_3, EXAMPLE_VALUE_3, EXAMPLE_VALUE_4, EXAMPLE_PROPERTY_4


class _Example(BaseModel):
    def __init__(self, example_1=None, example_2=None, example_3=None, *, example_4=None):
        self.example_1 = example_1
        self.example_2 = example_2
        self.example_3 = example_3
        self.example_4 = example_4


_example_mappings = [
    JsonPropertyMapping(EXAMPLE_PROPERTY_1, "example_1", parent_json_properties=["lots", "of", "nesting"]),
    JsonPropertyMapping(EXAMPLE_PROPERTY_2, "example_2", parent_json_properties=["lots", "of", "nesting"]),
    JsonPropertyMapping(json_property_getter=lambda obj_as_json: obj_as_json[EXAMPLE_PROPERTY_3],
                        json_property_setter=lambda obj_as_json, value: obj_as_json.update({EXAMPLE_PROPERTY_3: value}),
                        object_property_name="example_3", parent_json_properties=["more", "nesting"]),
    JsonPropertyMapping(EXAMPLE_PROPERTY_4, "example_4", parent_json_properties=["nesting"], optional=True),
]
_ExampleJSONEncoder = MappingJSONEncoderClassBuilder(_Example, _example_mappings).build()
_ExampleJSONDecoder = MappingJSONDecoderClassBuilder(_Example, _example_mappings).build()


_EXAMPLE = _Example(EXAMPLE_VALUE_1, EXAMPLE_VALUE_2, EXAMPLE_VALUE_3)
_EXAMPLE_AS_JSON = {
    "lots": {
        "of": {
            "nesting": {
                EXAMPLE_PROPERTY_1: EXAMPLE_VALUE_1,
                EXAMPLE_PROPERTY_2: EXAMPLE_VALUE_2
            }
        }
    },
    "more": {
        "nesting": {
            EXAMPLE_PROPERTY_3: EXAMPLE_VALUE_3
        }
    }
}

_EXAMPLE_WITH_OPTIONAL = _Example(EXAMPLE_VALUE_1, EXAMPLE_VALUE_2, EXAMPLE_VALUE_3, example_4=EXAMPLE_VALUE_4)
_EXAMPLE_WITH_OPTIONAL_AS_JSON = dict(**_EXAMPLE_AS_JSON, nesting={EXAMPLE_PROPERTY_4: EXAMPLE_VALUE_4})


class TestParentJsonProperties(unittest.TestCase):
    """
    Tests for ability to user `parent_json_properties`.
    """
    def test_encode(self):
        obj_as_dict = _ExampleJSONEncoder().default(_EXAMPLE)
        self.assertEqual(_EXAMPLE_AS_JSON, obj_as_dict)

    def test_encode_when_optional_is_defined(self):
        obj_as_dict = _ExampleJSONEncoder().default(_EXAMPLE_WITH_OPTIONAL)
        self.assertEqual(_EXAMPLE_WITH_OPTIONAL_AS_JSON, obj_as_dict)

    def test_decode(self):
        obj = _ExampleJSONDecoder().decode_parsed(_EXAMPLE_AS_JSON)
        self.assertEqual(_EXAMPLE, obj)

    def test_decode_when_optional_is_defined(self):
        obj = _ExampleJSONDecoder().decode_parsed(_EXAMPLE_WITH_OPTIONAL_AS_JSON)
        self.assertEqual(_EXAMPLE_WITH_OPTIONAL, obj)

    def test_decode_when_missing_nesting(self):
        example_as_json = deepcopy(_EXAMPLE_AS_JSON)
        del example_as_json["more"]["nesting"]
        self.assertRaises(KeyError, _ExampleJSONDecoder().decode_parsed, example_as_json)


if __name__ == "__main__":
    unittest.main()



import json
import unittest
from datetime import datetime, timezone

from hgijson.json_converters.primitive import StrJSONDecoder, IntJSONEncoder, FloatJSONEncoder, FloatJSONDecoder, \
    DatetimeEpochJSONEncoder, DatetimeEpochJSONDecoder, DatetimeISOFormatJSONDecoder, DatetimeISOFormatJSONEncoder, \
    IntJSONDecoder
from hgijson.json_converters.primitive import StrJSONEncoder


class TestStrJSONEncoder(unittest.TestCase):
    """
    Tests for `StrJSONEncoder`.
    """
    def test_default_with_str(self):
        self.assertEqual("123", StrJSONEncoder().default("123"))

    def test_default_with_int(self):
        self.assertEqual("123", StrJSONEncoder().default(123))

    def test_default_with_float(self):
        self.assertEqual("12.3", StrJSONEncoder().default(12.3))


class TestStrJSONDecoder(unittest.TestCase):
    """
    Tests for `TestStrJSONDecoder`.
    """
    def test_decode_with_string(self):
        self.assertEqual("123", StrJSONDecoder().decode("123"))

    def test_with_json_loads(self):
        self.assertEqual("123", json.loads("123", cls=StrJSONDecoder))


class TestIntJSONEncoder(unittest.TestCase):
    """
    Tests for `IntJSONEncoder`.
    """
    def test_default_with_str(self):
        self.assertEqual(123, IntJSONEncoder().default("123"))

    def test_default_with_int(self):
        self.assertEqual(123, IntJSONEncoder().default(123))

    def test_default_with_float(self):
        self.assertEqual(12, IntJSONEncoder().default(12.3))


class TestIntJSONDecoder(unittest.TestCase):
    """
    Tests for `IntJSONDecoder`.
    """
    def test_decode_with_int_as_string(self):
        self.assertEqual(123, IntJSONDecoder().decode("123"))

    def test_with_json_loads_and_int_as_string(self):
        self.assertEqual(123, json.loads("123", cls=IntJSONDecoder))


class TestFloatJSONEncoder(unittest.TestCase):
    """
    Tests for `FloatJSONEncoder`.
    """
    def test_default_with_str(self):
        self.assertEqual(12.3, FloatJSONEncoder().default("12.3"))

    def test_default_with_int(self):
        self.assertEqual(123.0, FloatJSONEncoder().default(123))

    def test_default_with_float(self):
        self.assertEqual(12.3, FloatJSONEncoder().default(12.3))


class TestFloatJSONDecoder(unittest.TestCase):
    """
    Tests for `FloatJSONDecoder`.
    """
    def test_decode_with_float_as_string(self):
        self.assertEqual(12.3, FloatJSONDecoder().decode("12.3"))

    def test_with_json_loads_and_int_as_string(self):
        self.assertEqual(12.3, json.loads("12.3", cls=FloatJSONDecoder))


class TestDatetimeEpochJSONEncoder(unittest.TestCase):
    """
    Tests for `DatetimeEpochJSONEncoder`.
    """
    def test_default(self):
        value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(0, DatetimeEpochJSONEncoder().default(value))


class TestDatetimeEpochJSONDecoder(unittest.TestCase):
    """
    Tests for `DatetimeEpochJSONDecoder`.
    """
    def test_decode(self):
        expected_value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(expected_value, DatetimeEpochJSONDecoder().decode("0"))

    def test_with_json_loads(self):
        expected_value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(expected_value, json.loads("0", cls=DatetimeEpochJSONDecoder))


class TestDatetimeISOFormatJSONEncoder(unittest.TestCase):
    """
    Tests for `DatetimeISOFormatJSONEncoder`.
    """
    def test_default(self):
        value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual("1970-01-01T00:00:00+00:00", DatetimeISOFormatJSONEncoder().default(value))


class TestDatetimeISOFormatJSONDecoder(unittest.TestCase):
    """
    Tests for `DatetimeISOFormatJSONDecoder`.
    """
    def test_decode_with_numerical_offset(self):
        expected_value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(expected_value, DatetimeISOFormatJSONDecoder().decode_parsed("1970-01-01T00:00:00+00:00"))

    def test_decode_with_special_representation_offset(self):
        expected_value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(expected_value, DatetimeISOFormatJSONDecoder().decode_parsed("1970-01-01T00:00:00Z"))

    def test_with_json_loads(self):
        expected_value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(expected_value, json.loads('"1970-01-01T00:00:00+00:00"', cls=DatetimeISOFormatJSONDecoder))


if __name__ == "__main__":
    unittest.main()



import json
import unittest

from hgijson.tests.json_converters._helpers import create_complex_model_with_json_representation, \
    create_simple_model_with_json_representation
from hgijson.tests.json_converters._serializers import SimpleModelMappingJSONDecoder, ComplexModelMappingJSONDecoder, \
    SimpleModelMappingJSONEncoder, ComplexModelMappingJSONEncoder


class TestMappingJSONEncoder(unittest.TestCase):
    """
    Tests for `MappingJSONEncoder`.
    """
    def setUp(self):
        self.simple_model = create_simple_model_with_json_representation()[0]
        self.simple_model_as_json = create_simple_model_with_json_representation()[1]
        self.complex_model = create_complex_model_with_json_representation()[0]
        self.complex_model_as_json = create_complex_model_with_json_representation()[1]

    def test_default_with_unknown(self):
        self.assertRaises(TypeError, SimpleModelMappingJSONEncoder().default, object())

    def test_default_with_none(self):
        encoded = SimpleModelMappingJSONEncoder().default(None)
        self.assertIsNone(encoded)

    def test_default_with_simple(self):
        encoded = SimpleModelMappingJSONEncoder().default(self.simple_model)
        self.assertDictEqual(encoded, self.simple_model_as_json)

    def test_default_with_complex(self):
        encoded = ComplexModelMappingJSONEncoder().default(self.complex_model)
        self.assertDictEqual(encoded, self.complex_model_as_json)

    def test_default_when_none_property(self):
        self.complex_model.d = None
        encoded = ComplexModelMappingJSONEncoder().default(self.complex_model)
        self.assertIsNone(encoded["serialized_d"])

    def test_default_with_iterable_with_no_items(self):
        encoded = SimpleModelMappingJSONEncoder().default([])
        self.assertCountEqual(encoded, [])

    def test_default_with_iterable(self):
        simple_models = [create_simple_model_with_json_representation(i)[0] for i in range(10)]
        simple_models_as_json = [create_simple_model_with_json_representation(i)[1] for i in range(10)]
        encoded = SimpleModelMappingJSONEncoder().default(simple_models)
        self.assertCountEqual(encoded, simple_models_as_json)

    def test_class_with_json_dumps(self):
        encoded = json.dumps(self.complex_model, cls=ComplexModelMappingJSONEncoder)
        encoded_as_dict = json.loads(encoded)
        self.assertDictEqual(encoded_as_dict, self.complex_model_as_json)

    def test_class_with_json_dumps_when_iterable(self):
        complex_models = [create_complex_model_with_json_representation(i)[0] for i in range(10)]
        complex_models_as_json = [create_complex_model_with_json_representation(i)[1] for i in range(10)]
        encoded = json.dumps(complex_models, cls=ComplexModelMappingJSONEncoder)
        encoded_as_dict = json.loads(encoded)
        self.assertCountEqual(encoded_as_dict, complex_models_as_json)


class TestMappingJSONDecoder(unittest.TestCase):
    """
    Tests for `MappingJSONDecoder`.
    """
    def setUp(self):
        self.simple_model = create_simple_model_with_json_representation()[0]
        self.simple_model_as_json = create_simple_model_with_json_representation()[1]
        self.complex_model = create_complex_model_with_json_representation()[0]
        self.complex_model_as_json = create_complex_model_with_json_representation()[1]

    def test_decode_with_malformed_json(self):
        self.assertRaises(ValueError, SimpleModelMappingJSONDecoder().decode, ":)")

    def test_decode_with_none(self):
        null_json_value = json.dumps(None)
        decoded = SimpleModelMappingJSONDecoder().decode(null_json_value)
        self.assertIsNone(decoded)

    def test_decode_with_simple(self):
        object_as_json_string = json.dumps(self.simple_model_as_json)
        decoded = SimpleModelMappingJSONDecoder().decode(object_as_json_string)
        self.assertEqual(decoded, self.simple_model)

    def test_decode_with_complex(self):
        object_as_json_string = json.dumps(self.complex_model_as_json)
        decoded = ComplexModelMappingJSONDecoder().decode(object_as_json_string)
        self.assertEqual(decoded, self.complex_model)

    def test_decode_when_none_property(self):
        self.complex_model_as_json["serialized_d"] = None
        object_as_json_string = json.dumps(self.complex_model_as_json)
        decoded = ComplexModelMappingJSONDecoder().decode(object_as_json_string)
        self.assertIsNone(decoded.d)

    def test_decode_with_iterable_with_no_items(self):
        json_as_string = json.dumps([])
        decoded = SimpleModelMappingJSONDecoder().decode(json_as_string)
        self.assertCountEqual(decoded, [])

    def test_decode_with_iterable(self):
        simple_models = [create_simple_model_with_json_representation(i)[0] for i in range(10)]
        simple_models_as_json = [create_simple_model_with_json_representation(i)[1] for i in range(10)]
        json_as_string = json.dumps(simple_models_as_json)
        decoded = SimpleModelMappingJSONDecoder().decode(json_as_string)
        self.assertCountEqual(decoded, simple_models)

    def test_class_with_json_loads_with_single(self):
        object_as_json_string = json.dumps(self.complex_model_as_json)
        decoded = json.loads(object_as_json_string, cls=ComplexModelMappingJSONDecoder)
        self.assertEqual(decoded, self.complex_model)

    def test_class_with_json_loads_with_iterable(self):
        complex_models = [create_complex_model_with_json_representation(i)[0] for i in range(10)]
        complex_models_as_json = [create_complex_model_with_json_representation(i)[1] for i in range(10)]
        json_as_string = json.dumps(complex_models_as_json)
        decoded = json.loads(json_as_string, cls=ComplexModelMappingJSONDecoder)
        self.assertEqual(decoded, complex_models)


if __name__ == "__main__":
    unittest.main()



import copy
from abc import ABCMeta, abstractstaticmethod
from json import JSONEncoder
from typing import Dict, Optional, Iterable, Any, TypeVar

from hgijson.custom_types import PrimitiveJsonType


class _RegisteredTypeJSONEncoder(JSONEncoder, metaclass=ABCMeta):
    """
    JSON encoder that will encode objects using the registered encoders. Works with in-built JSON library:
    ```
    import json

    register_json_encoder(MyObjectType, MyEncoderType)
    json.dumps(object_to_serialise, target_cls=_RegisteredTypeJSONEncoder)
    ```
    """
    @abstractstaticmethod
    def _get_json_encoders_for_type(type_to_encode: type) -> Optional[Iterable[JSONEncoder]]:
        """
        Gets the correct JSON encoder for the given type.
        :param type_to_encode: the type to encode
        :return: the encoder for the given type
        """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs
        self._encoder_cache = dict()    # type: Dict[type, JSONEncoder]

    def default(self, to_encode: Any) -> PrimitiveJsonType:
        type_to_encode = type(to_encode)

        encoder_type = self._get_json_encoders_for_type(type_to_encode)
        if encoder_type is None:
            # Unknown type: let standard JSON parser deal with it (will almost certainly raise an exception)
            encoder_type = JSONEncoder
        assert isinstance(encoder_type, type)

        if encoder_type not in self._encoder_cache:
            encoder = encoder_type(*self._args, **self._kwargs)
            self._encoder_cache[encoder_type] = encoder

        encoder = self._encoder_cache[encoder_type]
        assert isinstance(encoder, JSONEncoder)

        return encoder.default(to_encode)


RegisteredTypeJSONEncoderType = TypeVar("RegisteredTypeJSONEncoder", bound=_RegisteredTypeJSONEncoder)


class AutomaticJSONEncoderClassBuilder:
    """
    Builder for `JSONEncoder` class that is able to use a number of given `JSONEncoders` to automatically serialise
    models that may contain many models of different types.

    Similar to `jsonpublish` but without the global scope and able to be used with any json.dumps settings (unlike
    `jsonpublish` which (confusingly) can only be used by the settings that the developer made (see the package's
    __init__).
    """
    # Encoders for objects that are handled by the in-build JSON library
    _DEFAULT_JSON_ENCODERS = {
        dict: JSONEncoder,
        list: JSONEncoder,
        tuple: JSONEncoder,
        str: JSONEncoder,
        int: JSONEncoder,
        float: JSONEncoder,
        bool: JSONEncoder,
        type(None): JSONEncoder
    }

    def __init__(self):
        """
        Constructor.
        """
        self._json_encoders = dict()    # type: Dict[type, type]
        self.reset_registered_json_encoders()

    def get_json_encoders_for_type(self, type_to_encode: type) -> Optional[Iterable[JSONEncoder]]:
        """
        Gets the registered JSON encoder for the given type.
        :param type_to_encode: the type of object that is to be encoded
        :return: the encoder for the given object else `None` if unknown
        """
        if type_to_encode not in self._json_encoders:
            return None
        return self._json_encoders[type_to_encode]

    def register_json_encoder(self, encoder_type: type, encoder: JSONEncoder):
        """
        Register the given JSON encoder for use with the given object type.
        :param encoder_type: the type of object to encode
        :param encoder: the JSON encoder
        :return: this builder
        """
        self._json_encoders[encoder_type] = encoder
        return self

    def reset_registered_json_encoders(self):
        """
        Resets registered JSON encoders so that only ones supported by the in-built library are supported.
        """
        self._json_encoders = copy.copy(AutomaticJSONEncoderClassBuilder._DEFAULT_JSON_ENCODERS)

    def build(self) -> RegisteredTypeJSONEncoderType:
        """
        Builds JSON encoder that uses the encoders registered at the point in time when this method is called.
        :return: the JSON encoder
        """
        class_name = "%s_%s" % (_RegisteredTypeJSONEncoder.__class__.__name__, id(self))
        # Use encoders set at the point in time at which the encoder was built
        builder_snapshot = copy.deepcopy(self)
        return type(
                class_name,
                (_RegisteredTypeJSONEncoder, ),
                {
                    "_get_json_encoders_for_type": builder_snapshot.get_json_encoders_for_type
                }
        )



import unittest
from typing import Iterable, Any, Tuple, Callable, Optional, Dict

from hgijson import JsonPropertyMapping
from hgijson.serialization import PropertyMapping, Serializer, Deserializer
from hgijson.tests._models import SimpleModel, ComplexModel
from hgijson.tests._serializers import SimpleModelDeserializer, SimpleModelSerializer, ComplexModelSerializer, \
    ComplexModelDeserializer
from hgijson.tests.json_converters._helpers import create_complex_model_with_json_representation, \
    create_simple_model_with_json_representation
from hgijson.custom_types import PrimitiveJsonType


class _TestSerialization(unittest.TestCase):
    """
    TODO
    """
    def setUp(self):
        self.simple_model, self.simple_model_as_json = create_simple_model_with_json_representation()
        self.complex_model, self.complex_model_as_json = create_complex_model_with_json_representation()

    def test_serialization_with_no_mapping(self):
        mappings = [PropertyMapping()]
        self._assertSerialization(mappings, {}, SimpleModel())

    def test_serialization_using_names(self):
        mappings = [JsonPropertyMapping("serialized_a", "a"), JsonPropertyMapping("serialized_b", "b")]
        self._assertSerialization(mappings, self.simple_model_as_json, self.simple_model)

    def test_serialization_with_iterable_with_no_items(self):
        self._assertSerialization([], [], [])

    def test_serialization_with_iterable(self):
        simple_models = [create_simple_model_with_json_representation(i)[0] for i in range(10)]
        simple_models_as_json = [create_simple_model_with_json_representation(i)[1] for i in range(10)]
        self._assertSerialization(None, simple_models_as_json, simple_models)

    def test_serialization_optional_when_not_set(self):
        mappings = [JsonPropertyMapping("serialized_a", "a", optional=True),
                    JsonPropertyMapping("serialized_b", "b", optional=True)]
        self._assertSerialization(mappings, {}, SimpleModel())

    def test_serialization_optional_when_set(self):
        mappings = [JsonPropertyMapping("serialized_a", "a", optional=True),
                    JsonPropertyMapping("serialized_b", "b", optional=True)]
        self._assertSerialization(mappings, self.simple_model_as_json, self.simple_model)

    def test_serialization_none(self):
        self._assertSerialization(None, None, None)

    def test_serialization_when_property_does_not_exist(self):
        mappings = [JsonPropertyMapping("serialized_a", "z")]
        self.assertRaises(
            AttributeError, self._assertSerialization, mappings, self.simple_model_as_json, self.simple_model)

    def test_serialization_using_serialization_cls(self):
        mappings = [
            JsonPropertyMapping("serialized_a", "a"),
            JsonPropertyMapping("serialized_b", "b", "constructor_b"),
            JsonPropertyMapping("serialized_c", "c"),
            PropertyMapping(serialized_property_setter=lambda obj, value: obj.__setitem__("serialized_d", value),
                            object_property_getter=lambda obj: obj.d, serializer_cls=SimpleModelSerializer)]
        expected = {
            "serialized_a": self.complex_model.a,
            "serialized_b": self.complex_model.b,
            "serialized_c": self.complex_model.c,
            "serialized_d": [{
                "serialized_a": i,
                "serialized_b": self.complex_model.b + i
            } for i in range(len(self.complex_model.d))]}

        self._assertSerialization(
            mappings, expected, self.complex_model, ComplexModelSerializer, ComplexModelDeserializer)

    def test_serialization_collection(self):
        mappings = [JsonPropertyMapping("serialized_b", "b", "constructor_b"),
                    JsonPropertyMapping("serialized_i", "i", collection_factory=set)]
        expected_serialized = {"serialized_b": self.complex_model_as_json["serialized_b"],
                               "serialized_i": list(self.complex_model.i)}
        expected_deserialized = ComplexModel(self.complex_model.b)
        expected_deserialized.i = self.complex_model.i

        self._assertSerialization(
            mappings, expected_serialized, expected_deserialized, ComplexModelSerializer, ComplexModelDeserializer)

    def _assertSerialization(
            self, mappings: Optional[Iterable[PropertyMapping]], expected_serialized: PrimitiveJsonType,
            expected_deserialized: Any,
            serializer_factory: Callable[[Iterable[PropertyMapping]], Serializer]=SimpleModelSerializer,
            deserializer_factory: Callable[[Iterable[PropertyMapping]], Deserializer]=SimpleModelDeserializer) \
            -> Tuple[PrimitiveJsonType, Any]:
        # TODO: Switch (serialization/deserialization first) order depending on SUT

        serializer = serializer_factory(mappings)
        serialised = serializer.serialize(expected_deserialized)
        self.assertEqual(expected_serialized, serialised)

        deserializer = deserializer_factory(mappings)
        deserialised = deserializer.deserialize(serialised)
        self.assertEqual(expected_deserialized, deserialised)

        return serialised, deserialised


class TestSerializer(_TestSerialization):
    """
    Tests for `Serializer`.
    """
    def test_serialize_using_object_property_getter(self):
        mappings = [JsonPropertyMapping("serialized_b_plus_one", object_property_getter=lambda model: model.b + 1)]
        serializer = SimpleModelSerializer(mappings)
        self.assertDictEqual(serializer.serialize(self.simple_model), {"serialized_b_plus_one": self.simple_model.b + 1})

    def test_serialize_using_serialized_property_setter(self):
        def serialized_property_setter(container: Dict, value: Dict):
            container["augmented_key_%s" % value] = value

        mappings = [JsonPropertyMapping(None, "b", json_property_setter=serialized_property_setter)]
        serializer = ComplexModelSerializer(mappings)
        self.assertDictEqual(serializer.serialize(self.complex_model), {"augmented_key_5": 5})


class TestDeserializer(_TestSerialization):
    """
    Tests for `Deserializer`.
    """
    def test_deserialize_using_constructor_parameter(self):
        mappings = [JsonPropertyMapping("serialized_a", "a"),
                    JsonPropertyMapping("serialized_b", object_constructor_parameter_name="constructor_b")]
        deserializer = SimpleModelDeserializer(mappings)
        self.assertEqual(deserializer.deserialize(self.simple_model_as_json), self.simple_model)

    def test_deserialize_using_constructor_parameter_with_modifier(self):
        mappings = [JsonPropertyMapping("serialized_a", "a", object_constructor_parameter_name="constructor_b",
                                        object_constructor_argument_modifier=lambda value: value + 1)]
        deserializer = SimpleModelDeserializer(mappings)
        model = SimpleModel(self.simple_model_as_json["serialized_a"] + 1)
        self.assertEqual(deserializer.deserialize(self.simple_model_as_json), model)

    def test_deserialize_using_serialized_property_getter(self):
        a_and_b = {
            "ab": [self.simple_model.a, self.simple_model.b]
        }
        mappings = [JsonPropertyMapping(None, "a", json_property_getter=lambda json_as_dict: json_as_dict["ab"][0]),
                    JsonPropertyMapping(None, "b", json_property_getter=lambda json_as_dict: json_as_dict["ab"][1])]
        deserializer = SimpleModelDeserializer(mappings)
        self.assertEqual(deserializer.deserialize(a_and_b), self.simple_model)


if __name__ == "__main__":
    unittest.main()



from typing import Dict, Tuple

from hgijson.tests._models import SimpleModel, ComplexModel

EXAMPLE_VALUE_1 = "example-value-1"
EXAMPLE_VALUE_2 = "example-value-2"
EXAMPLE_VALUE_3 = "example-value-3"
EXAMPLE_VALUE_4 = "example-value-4"

EXAMPLE_PROPERTY_1 = "example-property-1"
EXAMPLE_PROPERTY_2 = "example-property-2"
EXAMPLE_PROPERTY_3 = "example-property-3"
EXAMPLE_PROPERTY_4 = "example-property-4"


def create_simple_model_with_json_representation(modifier: int=0) -> Tuple[SimpleModel, Dict]:
    """
    Creates an instance of `SimpleModel` alongside its expected JSON representation (given the most obvious property
    mappings).
    :param modifier: can be used to make the model distinguishable
    :return: tuple where the first value is the model and the second is its JSON representation
    """
    simple_model = SimpleModel(50)
    simple_model.a = modifier

    simple_model_as_json = {
        "serialized_a": simple_model.a,
        "serialized_b": simple_model.b
    }

    return simple_model, simple_model_as_json


def create_complex_model_with_json_representation(modifier: int=0) -> Tuple[ComplexModel, Dict]:
    """
    Creates an instance of `ComplexModel` alongside its expected JSON representation (given the most obvious property
    mappings).
    :param modifier: can be used to make the model distinguishable
    :return: tuple where the first value is the model and the second is its JSON representation
    """
    complex_model = ComplexModel(5)
    complex_model.a = modifier
    complex_model.c = 4

    complex_model_as_json = create_simple_model_with_json_representation(modifier)[1]
    complex_model_as_json.update({
        "serialized_a": complex_model.a,
        "serialized_b": complex_model.b,
        "serialized_c": complex_model.c,
        "serialized_d": [{
            "serialized_a": i,
            "serialized_b": complex_model.b + i
        } for i in range(3)],
        "serialized_e": complex_model.e,
        "serialized_f": complex_model.f,
        "serialized_g": complex_model.g,
        "serialized_h": complex_model.h,
        "serialized_i": list(complex_model.i)
    })

    return complex_model, complex_model_as_json



import unittest
from typing import Iterable, TypeVar, Generic, List, Set

from hgijson import JsonPropertyMapping, MappingJSONEncoderClassBuilder, MappingJSONDecoderClassBuilder
from hgijson.tests._models import BaseModel

_JSON_CUSTOM_COLLECTION_PROPERTY = "json-custom-collection"
_JSON_SET_COLLECTION_PROPERTY = "json-set-collection"

_CollectionItemType = TypeVar("CollectionItemType")


class _CustomCollection(Generic[_CollectionItemType], BaseModel):
    def __init__(self, items: Iterable[_CollectionItemType]=None):
        self._items = set(items) if items is not None else set()

    def add(self, item: _CollectionItemType):
        self._items.add(item)

    def get_iterator(self):
        return iter(self._items)


class _Container(BaseModel):
    @property
    def set_collection(self):
        return self._set_collection

    def __init__(self, set_collection: Set, custom_collection_items: List[int]=None):
        self.custom_collection = _CustomCollection[int](custom_collection_items)
        self._set_collection = set_collection


_container_mappings = [
    JsonPropertyMapping(_JSON_CUSTOM_COLLECTION_PROPERTY, "custom_collection", collection_factory=_CustomCollection,
                        collection_iter=lambda collection: collection.get_iterator()),
    JsonPropertyMapping(
        _JSON_SET_COLLECTION_PROPERTY, object_property_getter=lambda container: container.set_collection,
        object_constructor_parameter_name="set_collection", collection_factory=set),
]
_ContainerJSONEncoder = MappingJSONEncoderClassBuilder(_Container, _container_mappings).build()
_ContainerJSONDecoder = MappingJSONDecoderClassBuilder(_Container, _container_mappings).build()


class TestCollectionSupport(unittest.TestCase):
    """
    Testing for collection support.
    """
    def setUp(self):
        self.model = _Container({1, 2, 3}, [4, 5, 6])
        self.model_as_json = {
            _JSON_CUSTOM_COLLECTION_PROPERTY: list(self.model.custom_collection.get_iterator()),
            _JSON_SET_COLLECTION_PROPERTY: list(self.model.set_collection)
        }

    def test_encode(self):
        model_as_json = _ContainerJSONEncoder().default(self.model)
        self.assertEqual(self.model_as_json, model_as_json)

    def test_decode(self):
        model = _ContainerJSONDecoder().decode_parsed(self.model_as_json)
        self.assertEqual(self.model, model)


if __name__ == "__main__":
    unittest.main()
